# -*- coding: utf-8 -*-

import itertools
from abc import ABC, abstractmethod
from copy import deepcopy

from sudokustepper.grid import Grid


class SolverDelegate:
    def on_solver_step_complete(self, grid: Grid):
        """
        Called each time a step in the solver is completed.

        :param grid: a copy of the grid from the current step
        """
        pass

    def on_solver_solved(self):
        """
        Called if the solver finds a solution to the puzzle.
        """
        pass

    def on_solver_failed(self):
        """
        Called if the solver was unable to find a solution for the puzzle.
        """
        pass


class Solver(ABC):
    def __init__(self, grid: Grid, delegate=None):
        self.grid: Grid = grid
        self.step_history = []
        self.delegate: SolverDelegate = delegate

    def _step_complete(self):
        self.step_history.append(deepcopy(self.grid))

        if self.delegate is not None:
            grid_copy = deepcopy(self.grid)
            self.delegate.on_solver_step_complete(grid_copy)

    def _solved(self):
        if self.delegate is not None:
            self.delegate.on_solver_solved()

    def _failed(self):
        if self.delegate is not None:
            self.delegate.on_solver_failed()

    @property
    def num_steps(self):
        return len(self.step_history)

    @abstractmethod
    def solve(self) -> bool:
        """
        Attempts to find a solution for the Sudoku grid.

        :returns: True if a solution has been found, otherwise False.
        """
        pass


class NaiveSolver(Solver):
    def solve(self):
        success = False

        # Initialisation
        empty_cell_coords = self.grid.empty_cell_coords()
        all_possible_cell_values = []
        for (x, y) in empty_cell_coords:
            possible_values = self.grid.possible_values_for_cell(x, y)
            all_possible_cell_values.append(possible_values)

        # Solving
        for values_to_try in itertools.product(*all_possible_cell_values):
            for i, (x, y) in enumerate(empty_cell_coords):
                self.grid.cells[y][x].value = values_to_try[i]

            if self.grid.valid:
                success = True
                break

            self._step_complete()

        if success:
            self._solved()
        else:
            self._failed()

        return success


class BacktracingSolver(Solver):
    def solve(self):
        # Find the next empty cell
        empty_cell_coords = self.grid.empty_cell_coords()
        if not empty_cell_coords:
            self._failed()
            return False

        x, y = empty_cell_coords[0]
        for possible_value in self.grid.possible_values_for_cell(x, y):
            self.grid.cells[y][x].value = possible_value

            self._step_complete()

            if self.grid.solved:
                self._solved()
                return True
            else:
                if self.solve():
                    return True

        self.grid.cells[y][x].value = 0
        return False


ALL_SOLVERS = {
    "naive": NaiveSolver,
    "backtracing": BacktracingSolver
}


def main():
    g = Grid("862341950"
             "573000800"
             "900007023"
             "009510460"
             "080602070"
             "025034100"
             "240173589"
             "008000716"
             "050986234")

    print(g)
    print("Valid? {}".format(g.valid))
    print("Solved? {}".format(g.solved))

    s = NaiveSolver(g)
    if s.solve():
        print(g)
        print("Valid? {}".format(g.valid))
        print("Solved? {}".format(g.solved))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import threading
from copy import deepcopy
from typing import List

from PyQt5.QtCore import pyqtSignal, pyqtProperty, pyqtSlot, Qt, QSize, QTimer, QFile, QTextStream
from PyQt5.QtGui import QPaintEvent
from PyQt5.QtWidgets import *

from sudokustepper import solvers
from sudokustepper.grid import Cell, Grid


class CellWidget(QLabel):
    def __init__(self):
        super().__init__()

        self._cell: Cell = Cell()
        self._locked: bool = False
        self._valid: bool = True

        self.setAlignment(Qt.AlignCenter)

    def minimumSizeHint(self):
        return QSize(50, 50)

    @property
    def cell(self) -> Cell:
        return self._cell

    @cell.setter
    def cell(self, cell) -> None:
        self._cell = cell
        self.update()

    def paintEvent(self, e: QPaintEvent) -> None:
        self._locked = self._cell.locked
        self._valid = self._cell.valid

        t = "" if self.cell.empty else str(self.cell.value)
        self.setText(t)

        # Qt trick to reload the QSS style for this widget
        self.style().unpolish(self)
        self.style().polish(self)

        super().paintEvent(e)

    @pyqtProperty(bool)
    def valid(self) -> bool:
        return self._valid

    @pyqtProperty(bool)
    def locked(self) -> bool:
        return self._locked


class GridWidget(QWidget):
    def __init__(self):
        super().__init__()

        self._grid = Grid()
        self._cell_widgets = None
        self.init_ui()

    def init_ui(self):
        grid_layout = QGridLayout()
        grid_layout.setSpacing(0)
        self.setLayout(grid_layout)

        # 9x9 grid of buttons
        self._cell_widgets: List[List[CellWidget]] = []
        for i in range(9):
            row = []
            for j in range(9):
                b = CellWidget()
                b.cell = self._grid.cells[i][j]
                row.append(b)
                grid_layout.addWidget(b, i, j)
            self._cell_widgets.append(row)

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, grid):
        self._grid = grid
        for i in range(9):
            for j in range(9):
                self._cell_widgets[i][j].cell = self._grid.cells[i][j]
        self.update()
        

class PlaybackControlsWidget(QWidget):
    step_selected = pyqtSignal(int)

    def __init__(self):
        super().__init__()

        self._playback_timer = QTimer(self)
        self._playback_timer.timeout.connect(self._step)

        self._slider_steps = None
        self._lbl_end_step = None
        self._btn_play_pause = None
        self.init_ui()

    def init_ui(self):
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)

        #
        # Top layout - scrubbing slider
        #
        scrubbing_layout = QHBoxLayout()
        main_layout.addLayout(scrubbing_layout)

        lbl_start_step = QLabel("Original")

        self._slider_steps = QSlider(Qt.Horizontal)
        self._slider_steps.valueChanged.connect(self.step_selected)
        self._slider_steps.sliderMoved.connect(self.pause)  # If the user manually drags the slider, pause playback
        self._slider_steps.setMinimum(0)
        self._slider_steps.setSingleStep(1)

        self._lbl_end_step = QLabel("?")

        scrubbing_layout.addWidget(lbl_start_step)
        scrubbing_layout.addWidget(self._slider_steps)
        scrubbing_layout.addWidget(self._lbl_end_step)

        #
        # Bottom layout - playback control buttons
        #
        controls_layout = QHBoxLayout()
        main_layout.addLayout(controls_layout)

        btn_rewind = QPushButton("<<<")
        btn_rewind.clicked.connect(self.rewind)

        btn_step_back = QPushButton("<")
        btn_step_back.clicked.connect(self.step_back)

        self._btn_play_pause = QPushButton("Play")
        self._btn_play_pause.clicked.connect(self.toggle_play_pause)

        btn_step_next = QPushButton(">")
        btn_step_next.clicked.connect(self.step)
        btn_fast_forward = QPushButton(">>>")
        btn_fast_forward.clicked.connect(self.fastforward)

        controls_layout.addWidget(btn_rewind)
        controls_layout.addWidget(btn_step_back)
        controls_layout.addWidget(self._btn_play_pause)
        controls_layout.addWidget(btn_step_next)
        controls_layout.addWidget(btn_fast_forward)

    def reset(self, num_steps: int):
        self._slider_steps.setMaximum(num_steps)
        self._slider_steps.setSliderPosition(num_steps)
        self._lbl_end_step.setText("Step {}".format(num_steps))

    @pyqtSlot()
    def toggle_play_pause(self):
        self.pause() if self._playback_timer.isActive() else self.play()

    @property
    def play_head_at_start(self) -> bool:
        return self._slider_steps.sliderPosition() == self._slider_steps.minimum()

    @property
    def play_head_at_end(self) -> bool:
        return self._slider_steps.sliderPosition() == self._slider_steps.maximum()

    @pyqtSlot()
    def play(self):
        if self.play_head_at_end:
            self.rewind()

        self._playback_timer.start(100)
        self._btn_play_pause.setText("Pause")

    @pyqtSlot()
    def pause(self):
        self._playback_timer.stop()
        self._btn_play_pause.setText("Play")

    @pyqtSlot()
    def rewind(self):
        self.pause()
        self._slider_steps.setSliderPosition(self._slider_steps.minimum())

    @pyqtSlot()
    def fastforward(self):
        self.pause()
        self._slider_steps.setSliderPosition(self._slider_steps.maximum())

    @pyqtSlot()
    def step(self):
        self.pause()
        self._step()

    def _step(self):
        new_pos = self._slider_steps.sliderPosition() + 1
        self._slider_steps.setSliderPosition(new_pos)

        if self.play_head_at_end:
            self.pause()

    @pyqtSlot()
    def step_back(self):
        self.pause()
        new_pos = self._slider_steps.sliderPosition() - 1
        self._slider_steps.setSliderPosition(new_pos)


class LoadGridDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.grid: Grid = Grid()
        self._grid_preview = GridWidget()
        self._grid_preview.grid = self.grid

        self._line_edit_grid_string: QLineEdit = None
        self._button_box: QDialogButtonBox = None
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("Load a Grid")

        main_layout = QGridLayout()
        self.setLayout(main_layout)

        main_layout.addWidget(QLabel("Grid String"), 0, 0)
        self._line_edit_grid_string = QLineEdit()
        self._line_edit_grid_string.textChanged.connect(self.update_ui)
        main_layout.addWidget(self._line_edit_grid_string, 1, 0, alignment=Qt.AlignTop)

        main_layout.addWidget(QLabel("Preview"), 0, 1)
        main_layout.addWidget(self._grid_preview, 1, 1)

        self._button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)
        main_layout.addWidget(self._button_box, 2, 0, 1, 2)

        self.update_ui()

    @property
    def valid(self) -> bool:
        return self.grid.valid and not self.grid.solved

    @pyqtSlot()
    def update_ui(self) -> None:
        try:
            grid_str = self._line_edit_grid_string.text()
            self.grid.grid_string = grid_str

            self._grid_preview.update()
            self._button_box.button(QDialogButtonBox.Ok).setEnabled(self.valid)
        except ValueError:
            # TODO: notify the user of the error
            return


class SudokuSolverWindow(QMainWindow, solvers.SolverDelegate):
    def __init__(self):
        super().__init__()

        self.original_grid = Grid()
        self.solver: solvers.Solver = None
        self._solver_thread: threading.Thread = None

        self._grid_widget = None
        self._btn_load_grid = None
        self._combo_box_algorithm = None
        self._btn_start_solver = None
        self._playback_controls = None
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("Sudoku Solver")
        self.statusBar().showMessage("Load a grid to get started")

        main_layout = QVBoxLayout()
        w = QWidget()
        w.setLayout(main_layout)
        self.setCentralWidget(w)

        top_layout = QHBoxLayout()
        main_layout.addLayout(top_layout)

        # Left side - sudokustepper grid
        self._grid_widget = GridWidget()
        top_layout.addWidget(self._grid_widget)

        # Right side - options
        options_layout = QFormLayout()
        top_layout.addLayout(options_layout)

        self._btn_load_grid = QPushButton("Load Grid...")
        self._btn_load_grid.clicked.connect(self.load_grid_dialog)
        options_layout.addRow(self._btn_load_grid)

        divider_top = QFrame()
        divider_top.setFrameStyle(QFrame.HLine)
        divider_top.setFrameShadow(QFrame.Sunken)
        options_layout.addRow(divider_top)

        self._combo_box_algorithm = QComboBox()
        self._combo_box_algorithm.setEnabled(False)
        # Add all solver names to the list
        self._combo_box_algorithm.addItems([k.capitalize() for k in solvers.ALL_SOLVERS.keys()])
        self._combo_box_algorithm.setCurrentIndex(0)
        options_layout.addRow("Algorithm", self._combo_box_algorithm)

        self._btn_start_solver = QPushButton("Solve!")
        self._btn_start_solver.setEnabled(False)
        self._btn_start_solver.clicked.connect(self.start_solver)
        options_layout.addRow(self._btn_start_solver)

        self._playback_controls = PlaybackControlsWidget()
        self._playback_controls.setEnabled(False)
        self._playback_controls.step_selected.connect(self.preview_solver_step)
        main_layout.addWidget(self._playback_controls)

    def load_grid(self, grid: Grid):
        self.original_grid = deepcopy(grid)
        self._grid_widget.grid = self.original_grid
        self._grid_widget.update()

        self._combo_box_algorithm.setEnabled(True)
        self._btn_start_solver.setEnabled(True)
        self.statusBar().showMessage("Grid loaded")

    def on_solver_solved(self):
        self.statusBar().showMessage("Solved in {} seconds".format("..."))
        self._grid_widget.grid = self.solver.grid
        self._btn_load_grid.setEnabled(True)
        self._combo_box_algorithm.setEnabled(True)
        self._btn_start_solver.setEnabled(True)
        self._playback_controls.reset(self.solver.num_steps)
        self._playback_controls.setEnabled(True)

    def on_solver_failed(self):
        self._btn_load_grid.setEnabled(True)
        self._combo_box_algorithm.setEnabled(True)
        self._btn_start_solver.setEnabled(True)

    @pyqtSlot()
    def load_grid_dialog(self):
        dialog = LoadGridDialog(self)
        if dialog.exec():
            self.load_grid(dialog.grid)

    @pyqtSlot()
    def start_solver(self):
        self._btn_load_grid.setEnabled(False)
        self._combo_box_algorithm.setEnabled(False)
        self._btn_start_solver.setEnabled(False)

        solver_cls = solvers.ALL_SOLVERS[self._combo_box_algorithm.currentText().lower()]
        assert solver_cls is not None
        self.solver = solver_cls(deepcopy(self.original_grid), delegate=self)

        self._solver_thread = threading.Thread(target=self.solver.solve)
        self._solver_thread.start()

    @pyqtSlot(int)
    def preview_solver_step(self, step: int):
        if step == 0:
            self._grid_widget.grid = self.original_grid
            self.statusBar().showMessage("Original puzzle")
        else:
            self._grid_widget.grid = self.solver.step_history[step - 1]
            self.statusBar().showMessage("Showing step {}".format(step))


def main():
    import sys
    from sudokustepper import resources

    app = QApplication(sys.argv)

    # Load app stylesheet from the resource file
    style_sheet_file = QFile(":/styles/app.qss")
    style_sheet_file.open(QFile.ReadOnly | QFile.Text)
    style_sheet = QTextStream(style_sheet_file).readAll()
    app.setStyleSheet(style_sheet)

    w = SudokuSolverWindow()
    w.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from typing import List


class Cell:
    def __init__(self, value=0, locked=False):
        """
        Creates a cell with a specific value.
        :param value: must be between 0 and 9 inclusive
        :param locked: set to True if the cell value should be read-only, unless the value is 0 (empty), in which case
                       the cell cannot be locked and this value will be ignored
        """
        if value not in range(10):
            raise ValueError("cell value must be between 0 and 9 inclusive")

        self._value = value
        self.valid: bool = True

        # Only lock the cell if the value is non-zero
        if value == 0:
            self._locked = False
        else:
            self._locked = locked

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not self._locked and value in range(10):
            self._value = value

    @property
    def empty(self):
        return self.value == 0

    @property
    def locked(self):
        return self._locked

    def lock(self):
        if not self.empty:
            self._locked = True

    def unlock(self):
        self._locked = False

    def __eq__(self, o):
        if isinstance(o, Cell):
            return self.value == o.value
        return False

    def __hash__(self):
        return self.value

    def __str__(self):
        return " " if self.value == 0 else str(self.value)

    def __repr__(self):
        return "<Cell value:%d>" % self.value


class Grid:
    def __init__(self, grid_string: str = None):
        cells: List[List[Cell]] = []
        for i in range(9):
            row = []
            for j in range(9):
                row.append(Cell(0))
            cells.append(row)

        self.cells = cells
        self._valid = True

        if grid_string is not None:
            self.grid_string = grid_string

    @property
    def grid_string(self) -> str:
        result = []
        for cell in self.flattened():
            result.append(str(cell.value))

        return "".join(result)

    @grid_string.setter
    def grid_string(self, string: str):
        if len(string) != 81:
            raise ValueError("length of values must be 81")

        for i, cell in enumerate(self.flattened()):
            cell.unlock()
            cell.value = int(string[i])
            cell.lock()

        self.validate()

    def row(self, i) -> []:
        """
        Returns cells in the row at index i.
        :param i: the row index, counting from top to bottom
        :returns: a list of 9 cells
        """
        if i not in range(9):
            raise ValueError("i must be between 0 and 8 inclusive")
        row = self.cells[i]
        assert len(row) == 9

        return row

    def col(self, i) -> [Cell]:
        """
        Returns cells in the column at index i.

        :param i: the row index, counting from left to right

        :returns: a list of 9 cells
        """
        if i not in range(9):
            raise ValueError("i must be between 0 and 8 inclusive")
        col = [row[i] for row in self.rows()]
        assert len(col) == 9

        return col

    def box(self, i) -> [Cell]:
        """
        Returns cells in the box at index i.

        :param i: the box index, counting across from the top-left to
                  bottom-right

        :returns: a list of 9 cells, ordered across from top-left to
                  bottom-right
        """
        if i not in range(9):
            raise ValueError("i must be between 0 and 8 inclusive")
        row_start = 3 * (i // 3)
        col_start = 3 * (i % 3)
        box = []
        for row_i in range(row_start, row_start + 3):
            for col_i in range(col_start, col_start + 3):
                box.append(self.cells[row_i][col_i])

        return box

    def rows(self) -> [[Cell]]:
        """
        Returns a list of rows in the grid, ordered from top to bottom.

        :returns: a list of 9 rows, where each row is a list of 9 cells
        """
        rows = []
        for i in range(9):
            rows.append(self.row(i))

        return rows

    def cols(self) -> [[Cell]]:
        """
        Returns a list of columns in the grid, ordered from left to right.

        :returns: a list of 9 columns, where each column is a list of 9 cells
        """
        cols = []
        for i in range(9):
            cols.append(self.col(i))

        return cols

    def boxes(self) -> [[Cell]]:
        """
        Returns a list of 3x3 boxes in the grid, ordered from left-to-right
        then top-to-bottom

        :returns: a list of 9 boxes, where each box is a list of 9 cells
        """
        boxes = []
        for i in range(9):
            boxes.append(self.box(i))

        return boxes

    def flattened(self):
        return [cell for row in self.cells for cell in row]

    def empty_cell_coords(self) -> [(int, int)]:
        """
        Finds all empty cells in the grid.

        :returns: a list of coordinate tuples (x, y) of the grid's empty cells
        """
        coords = []
        for i in range(9):
            for j in range(9):
                if self.cells[i][j].empty:
                    coords.append((j, i))

        return coords

    @property
    def valid(self) -> bool:
        """
        :returns: True if the board is valid, otherwise False
        """
        self.validate()
        return self._valid

    def validate(self):
        """
        Validates cell values in the board, and sets the valid property to True if valid, otherwise False. Empty cells
        are ignored, thus the board is valid if empty.

        The board is valid if all of the following conditions are true:
            * each column contains the numbers 1-9 or blank cells, with no
              repeated values
            * each row follows the same rule
            * each 3x3 box follows the same rule
        """
        # Reset valid flag of all cells
        for cell in self.flattened():
            cell.valid = True

        # Get all cell groups, and remove the empty cells (we don't care about these)
        cell_groups = self.rows() + self.cols() + self.boxes()
        cell_groups = [
            [c for c in cells if not c.empty] for cells in cell_groups
        ]

        valid = True
        for cell_group in cell_groups:
            # Flag the duplicate cells in this cell group
            # TODO: make this algorithm more efficient
            seen_cells = set()
            for cell in cell_group:
                if cell in seen_cells:
                    valid &= False
                    cell.valid = False
                    # Also set the 'seen' cell as invalid
                    for c in seen_cells:
                        if c == cell:
                            c.valid = False
                            break
                seen_cells.add(cell)

        self._valid = valid

    @property
    def solved(self) -> bool:
        """
        Ensures all cells are populated, before checking if the grid is valid.

        :returns: True if the grid is solved, otherwise False
        """
        for row in self.cells:
            for cell in row:
                if cell.empty:
                    return False
        return self.valid

    def possible_values_for_cell(self, x: int, y: int) -> set:
        """
        Returns a set of the possible values for a specific cell.

        :param x: the cell's x coordinate, between 0 and 8 inclusive
        :param y: the cell's y coordinate, between 0 and 8 inclusive

        :returns: a set of possible values for the cell at (x, y)
        """
        row_values = set([cell.value for cell in self.row(y)])
        col_values = set([cell.value for cell in self.col(x)])
        box_index = 3 * (y // 3) + x // 3
        box_values = set([cell.value for cell in self.box(box_index)])

        values = set(range(1, 10)) - row_values - col_values - box_values
        return values

    def __eq__(self, o):
        if isinstance(o, Grid):
            return self.cells == o.cells
        return False

    def __str__(self):
        s = ""
        for i, row in enumerate(self.cells):
            if i == 0:
                s += "┌" + "─" * 9 + "┬" + "─" * 9 + "┬" + "─" * 9 + "┐\n"
            elif i % 3 == 0:
                s += "├" + "─" * 9 + "┼" + "─" * 9 + "┼" + "─" * 9 + "┤\n"
            for j, cell in enumerate(row):
                if j % 3 == 0:
                    s += "│"
                s += f" {cell} "
            s += "│\n"
        s += "└" + "─" * 9 + "┴" + "─" * 9 + "┴" + "─" * 9 + "┘"

        return s


if __name__ == "__main__":
    g = Grid("123456789" * 9)
    print(g)



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sudokustepper",
    version="0.2.1",
    author="Douglas Finlay",
    author_email="douglas@douglasfinlay.com",
    description="A tool for stepping through, and visualising Sudoku solving algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dougfinl/sudokustepper",
    packages=["sudokustepper"],
    python_requires=">=3",
    install_requires=["PyQt5"],
    entry_points={
        "gui_scripts": [
            "sudokustepper = sudokustepper.__main__:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Education",
        "Topic :: Games/Entertainment :: Puzzle Games",
        "Programming Language :: Python :: 3",
    ],
)
